import uuid
import weakref

from matplotlib.colors import ColorConverter

from glue.core.data import Subset, Data
from glue.core.exceptions import IncompatibleAttribute
from glue.utils import broadcast_to
from glue.core.fixed_resolution_buffer import ARRAY_CACHE, PIXEL_CACHE
from .colors import get_translucent_cmap
from .layer_state import VolumeLayerState
from ..common.layer_artist import VispyLayerArtist


class DataProxy(object):

    def __init__(self, viewer_state, layer_artist):
        self._viewer_state = weakref.ref(viewer_state)
        self._layer_artist = weakref.ref(layer_artist)

    @property
    def layer_artist(self):
        return self._layer_artist()

    @property
    def viewer_state(self):
        return self._viewer_state()

    @property
    def shape(self):

        x_axis = self.viewer_state.x_att.axis
        y_axis = self.viewer_state.y_att.axis
        z_axis = self.viewer_state.z_att.axis

        if isinstance(self.layer_artist.layer, Subset):
            full_shape = self.layer_artist.layer.data.shape
        else:
            full_shape = self.layer_artist.layer.shape

        return full_shape[z_axis], full_shape[y_axis], full_shape[x_axis]

    def compute_fixed_resolution_buffer(self, bounds=None):

        shape = [bound[2] for bound in bounds]

        if self.layer_artist is None or self.viewer_state is None:
            return broadcast_to(0, shape)

        if isinstance(self.layer_artist.layer, Subset):
            try:
                subset_state = self.layer_artist.layer.subset_state
                result = self.layer_artist.layer.data.compute_fixed_resolution_buffer(
                    target_data=self.layer_artist._viewer_state.reference_data,
                    bounds=bounds, subset_state=subset_state,
                    cache_id=self.layer_artist.id)
            except IncompatibleAttribute:
                self.layer_artist.disable_incompatible_subset()
                return broadcast_to(0, shape)
            else:
                self.layer_artist.enable()
        else:
            try:
                result = self.layer_artist.layer.compute_fixed_resolution_buffer(
                    target_data=self.layer_artist._viewer_state.reference_data,
                    bounds=bounds, target_cid=self.layer_artist.state.attribute,
                    cache_id=self.layer_artist.id)
            except IncompatibleAttribute:
                self.layer_artist.disable('Layer data is not fully linked to reference data')
                return broadcast_to(0, shape)
            else:
                self.layer_artist.enable()

        return result


class VolumeLayerArtist(VispyLayerArtist):
    """
    A layer artist to render volumes.

    This is more complex than for other visual types, because for volumes, we
    need to manage all the volumes via a single MultiVolume visual class for
    each data viewer.
    """

    def __init__(self, vispy_viewer=None, layer=None, layer_state=None):

        super(VolumeLayerArtist, self).__init__(layer)

        self._clip_limits = None

        self.layer = layer or layer_state.layer
        self.vispy_widget = vispy_viewer._vispy_widget

        # TODO: need to remove layers when layer artist is removed
        self._viewer_state = vispy_viewer.state
        self.state = layer_state or VolumeLayerState(layer=self.layer)
        if self.state not in self._viewer_state.layers:
            self._viewer_state.layers.append(self.state)

        # We create a unique ID for this layer artist, that will be used to
        # refer to the layer artist in the MultiVolume. We have to do this
        # rather than use self.id because we can't guarantee the latter is
        # unique.
        self.id = str(uuid.uuid4())

        self._multivol = self.vispy_widget._multivol
        self._multivol.allocate(self.id)

        self._viewer_state.add_global_callback(self._update_volume)
        self.state.add_global_callback(self._update_volume)

        self.reset_cache()

        self._data_proxy = None

    def reset_cache(self):
        self._last_viewer_state = {}
        self._last_layer_state = {}

    @property
    def visual(self):
        return self._multivol

    @property
    def bbox(self):
        return (-0.5, self.layer.shape[2] - 0.5,
                -0.5, self.layer.shape[1] - 0.5,
                -0.5, self.layer.shape[0] - 0.5)

    @property
    def shape(self):
        return self.layer.shape

    def redraw(self):
        """
        Redraw the Vispy canvas
        """
        self.vispy_widget.canvas.update()

    def clear(self):
        """
        Remove the layer artist from the visualization
        """
        # We don't want to deallocate here because this can be called if we
        # disable the layer due to incompatible attributes
        self._multivol.disable(self.id)

    def remove(self):
        """
        Remove the layer artist for good
        """
        self._multivol.deallocate(self.id)
        ARRAY_CACHE.pop(self.id, None)
        PIXEL_CACHE.pop(self.id, None)

    def _update_cmap_from_color(self):
        cmap = get_translucent_cmap(*ColorConverter().to_rgb(self.state.color))
        self._multivol.set_cmap(self.id, cmap)
        self.redraw()

    def _update_limits(self):
        if isinstance(self.layer, Subset):
            self._multivol.set_clim(self.id, None)
        else:
            self._multivol.set_clim(self.id, (self.state.vmin, self.state.vmax))
        self.redraw()

    def _update_alpha(self):
        self._multivol.set_weight(self.id, self.state.alpha)
        self.redraw()

    def _update_subset_mode(self):
        if isinstance(self.state.layer, Data) or self.state.subset_mode == 'outline':
            self._multivol.set_multiply(self.id, None)
        else:
            label = self._multivol.label_for_layer(self.state.layer.data)
            self._multivol.set_multiply(self.id, label)
        self.redraw()

    def _update_data(self):

        if self._data_proxy is None:
            self._data_proxy = DataProxy(self._viewer_state, self)
            self._multivol.set_data(self.id, self._data_proxy, layer=self.layer)
        else:
            self._multivol._update_scaled_data(self.id)

        self._update_subset_mode()

    def _update_visibility(self):
        if self.state.visible:
            self._multivol.enable(self.id)
        else:
            self._multivol.disable(self.id)
        self.redraw()

    def set_clip(self, limits):
        pass

    def _update_volume(self, force=False, **kwargs):

        if self.state.attribute is None or self.state.layer is None:
            return

        # Figure out which attributes are different from before. Ideally we shouldn't
        # need this but currently this method is called multiple times if an
        # attribute is changed due to x_att changing then hist_x_min, hist_x_max, etc.
        # If we can solve this so that _update_histogram is really only called once
        # then we could consider simplifying this. Until then, we manually keep track
        # of which properties have changed.

        changed = set()

        if not force:

            for key, value in self._viewer_state.as_dict().items():
                if value != self._last_viewer_state.get(key, None):
                    changed.add(key)

            for key, value in self.state.as_dict().items():
                if value != self._last_layer_state.get(key, None):
                    changed.add(key)

        self._last_viewer_state.update(self._viewer_state.as_dict())
        self._last_layer_state.update(self.state.as_dict())

        if force or 'color' in changed:
            self._update_cmap_from_color()

        if force or 'vmin' in changed or 'vmax' in changed:
            self._update_limits()

        if force or 'alpha' in changed:
            self._update_alpha()

        if force or 'layer' in changed or 'attribute' in changed:
            self._update_data()

        if force or 'subset_mode' in changed:
            self._update_subset_mode()

        if force or 'visible' in changed:
            self._update_visibility()

    def update(self):
        self._update_volume(force=True)
        self.redraw()



from glue.core.layer_artist import LayerArtistBase

__all__ = ['VispyLayerArtist']


class VispyLayerArtist(LayerArtistBase):

    @property
    def zorder(self):
        return self.state.zorder

    @zorder.setter
    def zorder(self, value):
        self.state.zorder = value
        self.redraw()  # TODO: determine if this is needed

    @property
    def visible(self):
        return self.state.visible

    @visible.setter
    def visible(self, value):
        self.state.visible = value

    def __gluestate__(self, context):
        return dict(state=context.id(self.state))



from glue.config import colormaps
from echo import CallbackProperty, SelectionCallbackProperty, delay_callback
from glue.core.state_objects import StateAttributeLimitsHelper
from glue.core.data_combo_helper import ComponentIDComboHelper

from ..common.layer_state import VispyLayerState


__all__ = ['IsosurfaceLayerState']


class IsosurfaceLayerState(VispyLayerState):
    """
    A state object for volume layers
    """

    attribute = SelectionCallbackProperty()

    level_low = CallbackProperty()
    level_high = CallbackProperty()
    cmap = CallbackProperty()
    step = CallbackProperty(4)

    level_cache = CallbackProperty({})

    def __init__(self, layer=None, **kwargs):

        super(IsosurfaceLayerState, self).__init__(layer=layer)

        self.att_helper = ComponentIDComboHelper(self, 'attribute')

        self.lim_helper = StateAttributeLimitsHelper(self, attribute='attribute',
                                                     lower='level_low', upper='level_high')

        self.add_callback('layer', self._on_layer_change)
        if layer is not None:
            self._on_layer_change()

        self.cmap = colormaps.members[0][1]

        self.update_from_dict(kwargs)

    def _on_layer_change(self, layer=None):

        with delay_callback(self, 'level_low', 'level_high'):

            if self.layer is None:
                self.att_helper.set_multiple_data([])
            else:
                self.att_helper.set_multiple_data([self.layer])

    def update_priority(self, name):
        return 0 if name == 'level' else 1



from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = 'undefined'

try:
    import OpenGL  # noqa
except ImportError:
    raise ImportError("The PyOpenGL package is required for this plugin")
else:
    del OpenGL



from glue.config import colormaps
from echo import CallbackProperty, SelectionCallbackProperty, keep_in_sync, delay_callback
from glue.core.state_objects import StateAttributeLimitsHelper
from glue.core.data_combo_helper import ComponentIDComboHelper
from ..common.layer_state import VispyLayerState

__all__ = ['ScatterLayerState']


class ScatterLayerState(VispyLayerState):
    """
    A state object for volume layers
    """

    size_mode = CallbackProperty('Fixed')
    size = CallbackProperty()
    size_attribute = SelectionCallbackProperty()
    size_vmin = CallbackProperty()
    size_vmax = CallbackProperty()
    size_scaling = CallbackProperty(1)

    color_mode = CallbackProperty('Fixed')
    cmap_attribute = SelectionCallbackProperty()
    cmap_vmin = CallbackProperty()
    cmap_vmax = CallbackProperty()
    cmap = CallbackProperty()

    xerr_visible = CallbackProperty(False)
    xerr_attribute = SelectionCallbackProperty()
    yerr_visible = CallbackProperty(False)
    yerr_attribute = SelectionCallbackProperty()
    zerr_visible = CallbackProperty(False)
    zerr_attribute = SelectionCallbackProperty()

    vector_visible = CallbackProperty(False)
    vx_attribute = SelectionCallbackProperty()
    vy_attribute = SelectionCallbackProperty()
    vz_attribute = SelectionCallbackProperty()
    vector_scaling = CallbackProperty(1)
    vector_origin = SelectionCallbackProperty(default_index=1)
    vector_arrowhead = CallbackProperty()

    size_limits_cache = CallbackProperty({})
    cmap_limits_cache = CallbackProperty({})

    def __init__(self, layer=None, **kwargs):

        self._sync_markersize = None

        super(ScatterLayerState, self).__init__(layer=layer)

        if self.layer is not None:

            self.color = self.layer.style.color
            self.size = self.layer.style.markersize
            self.alpha = self.layer.style.alpha

        self.size_att_helper = ComponentIDComboHelper(self, 'size_attribute')
        self.cmap_att_helper = ComponentIDComboHelper(self, 'cmap_attribute')
        self.xerr_att_helper = ComponentIDComboHelper(self, 'xerr_attribute', categorical=False)
        self.yerr_att_helper = ComponentIDComboHelper(self, 'yerr_attribute', categorical=False)
        self.zerr_att_helper = ComponentIDComboHelper(self, 'zerr_attribute', categorical=False)

        self.vx_att_helper = ComponentIDComboHelper(self, 'vx_attribute', categorical=False)
        self.vy_att_helper = ComponentIDComboHelper(self, 'vy_attribute', categorical=False)
        self.vz_att_helper = ComponentIDComboHelper(self, 'vz_attribute', categorical=False)

        self.size_lim_helper = StateAttributeLimitsHelper(self, attribute='size_attribute',
                                                          lower='size_vmin', upper='size_vmax',
                                                          cache=self.size_limits_cache)

        self.cmap_lim_helper = StateAttributeLimitsHelper(self, attribute='cmap_attribute',
                                                          lower='cmap_vmin', upper='cmap_vmax',
                                                          cache=self.cmap_limits_cache)

        vector_origin_display = {'tail': 'Tail of vector',
                                 'middle': 'Middle of vector',
                                 'tip': 'Tip of vector'}
        ScatterLayerState.vector_origin.set_choices(self, ['tail', 'middle', 'tip'])
        ScatterLayerState.vector_origin.set_display_func(self, vector_origin_display.get)

        self.add_callback('layer', self._on_layer_change)
        if layer is not None:
            self._on_layer_change()

        self.cmap = colormaps.members[0][1]

        self.update_from_dict(kwargs)

    def _on_layer_change(self, layer=None):

        with delay_callback(self, 'cmap_vmin', 'cmap_vmax', 'size_vmin', 'size_vmax'):
            helpers = [self.size_att_helper, self.cmap_att_helper,
                       self.xerr_att_helper, self.yerr_att_helper, self.zerr_att_helper,
                       self.vx_att_helper, self.vy_att_helper, self.vz_att_helper]
            if self.layer is None:
                for helper in helpers:
                    helper.set_multiple_data([])
            else:
                for helper in helpers:
                    helper.set_multiple_data([self.layer])

    def update_priority(self, name):
        return 0 if name.endswith(('vmin', 'vmax')) else 1

    def _layer_changed(self):

        super(ScatterLayerState, self)._layer_changed()

        if self._sync_markersize is not None:
            self._sync_markersize.stop_syncing()

        if self.layer is not None:
            self.size = self.layer.style.markersize
            self._sync_markersize = keep_in_sync(self, 'size', self.layer.style, 'markersize')

    def flip_size(self):
        self.size_lim_helper.flip_limits()

    def flip_cmap(self):
        self.cmap_lim_helper.flip_limits()



from ..common.vispy_data_viewer import BaseVispyViewer
from .layer_artist import ScatterLayerArtist
from .layer_style_widget import ScatterLayerStyleWidget
from .viewer_state import Vispy3DScatterViewerState

from ..common import tools as _tools, selection_tools  # noqa
from . import scatter_toolbar  # noqa


class VispyScatterViewer(BaseVispyViewer):

    LABEL = "3D Scatter"

    _state_cls = Vispy3DScatterViewerState
    _layer_style_widget_cls = ScatterLayerStyleWidget

    tools = BaseVispyViewer.tools + ['vispy:lasso', 'vispy:rectangle',
                                     'vispy:circle', 'scatter3d:point']

    _data_artist_cls = ScatterLayerArtist
    _subset_artist_cls = ScatterLayerArtist

    def add_data(self, data):

        first_layer_artist = len(self._layer_artist_container) == 0

        added = super(VispyScatterViewer, self).add_data(data)

        if added:
            if first_layer_artist:
                # The above call to add_data may have added subset layers, some
                # of which may be incompatible with the data, so we need to now
                # explicitly use the layer for the actual data object.
                layer = self._layer_artist_container[data][0]
                self.state.set_limits(*layer.default_limits)
                self._ready_draw = True

        return added



import numpy as np

from vispy.visuals.transforms import (ChainTransform, NullTransform,
                                      MatrixTransform, STTransform)
from vispy.visuals.transforms.base_transform import InverseTransform
from vispy.visuals.transforms._util import arg_to_vec4


def as_matrix_transform(transform):
    """
    Simplify a transform to a single matrix transform, which makes it a lot
    faster to compute transformations.

    Raises a TypeError if the transform cannot be simplified.
    """
    if isinstance(transform, ChainTransform):
        matrix = np.identity(4)
        for tr in transform.transforms:
            # We need to do the matrix multiplication manually because VisPy
            # somehow doesn't mutliply matrices if there is a perspective
            # component. The equation below looks like it's the wrong way
            # around, but the VisPy matrices are transposed.
            matrix = np.matmul(as_matrix_transform(tr).matrix, matrix)
        return MatrixTransform(matrix)
    elif isinstance(transform, InverseTransform):
        matrix = as_matrix_transform(transform._inverse)
        return MatrixTransform(matrix.inv_matrix)
    elif isinstance(transform, NullTransform):
        return MatrixTransform()
    elif isinstance(transform, STTransform):
        return transform.as_matrix()
    elif isinstance(transform, MatrixTransform):
        return transform
    else:
        raise TypeError("Could not simplify transform of type {0}".format(type(transform)))


try:

    from glue.utils.qt import fix_tab_widget_fontsize  # noqa

except ImportError:

    import platform
    from glue.utils.qt import get_qapp

    def fix_tab_widget_fontsize(tab_widget):
        """
        Because of a bug in Qt, tab titles on MacOS X don't have the right font size
        """
        if platform.system() == 'Darwin':
            app = get_qapp()
            app_font = app.font()
            tab_widget.setStyleSheet('font-size: {0}px'.format(app_font.pointSize()))


class NestedSTTransform(STTransform):

    glsl_map = """
        vec4 st_transform_map(vec4 pos) {
            return vec4((pos.xyz * $innerscale.xyz + $innertranslate.xyz * pos.w).xyz
                         * $scale.xyz + $translate.xyz * pos.w, pos.w);
        }
    """

    glsl_imap = """
        vec4 st_transform_imap(vec4 pos) {
            return vec4((((pos.xyz - $innertranslate.xyz * pos.w) / $innerscale.xyz)
                         - $translate.xyz * pos.w) / $scale.xyz, pos.w);
        }
    """

    def __init__(self):
        self.inner = STTransform()
        super(NestedSTTransform, self).__init__()

    @arg_to_vec4
    def map(self, coords):
        coords = self.inner.map(coords)
        coords = super(NestedSTTransform, self).map(coords)
        return coords

    @arg_to_vec4
    def imap(self, coords):
        coords = super(NestedSTTransform, self).imap(coords)
        coords = self.inner.imap(coords)
        return coords

    def _update_shaders(self):
        self._shader_map['scale'] = self.scale
        self._shader_map['translate'] = self.translate
        self._shader_imap['scale'] = self.scale
        self._shader_imap['translate'] = self.translate
        self._shader_map['innerscale'] = self.inner.scale
        self._shader_map['innertranslate'] = self.inner.translate
        self._shader_imap['innerscale'] = self.inner.scale
        self._shader_imap['innertranslate'] = self.inner.translate



from contextlib import contextmanager

import numpy as np

from matplotlib.colors import ColorConverter

from vispy import scene
from vispy.scene.visuals import Arrow


class MultiColorScatter(scene.visuals.Markers):
    """
    This is a helper class to make it easier to show multiple markers at
    specific positions and control exactly which marker should be on top of
    which.
    """

    def __init__(self, *args, **kwargs):
        self.layers = {}
        self._combined_data = None
        self._skip_update = False
        self._error_vector_widget = None
        super(MultiColorScatter, self).__init__(*args, **kwargs)

    @contextmanager
    def delay_update(self):
        self._skip_update = True
        yield
        self._skip_update = False

    def allocate(self, label):
        if label in self.layers:
            raise ValueError("Layer {0} already exists".format(label))
        else:
            self.layers[label] = {'data': None,
                                  'mask': None,
                                  'errors': None,
                                  'vectors': None,
                                  'draw_arrows': False,
                                  'color': np.asarray((1., 1., 1.)),
                                  'alpha': 1.,
                                  'zorder': lambda: 0,
                                  'size': 10,
                                  'visible': True}

    def deallocate(self, label):
        self.layers.pop(label)
        self._update()

    def set_data_values(self, label, x, y, z):
        """
        Set the position of the datapoints
        """
        # TODO: avoid re-allocating an array every time
        self.layers[label]['data'] = np.array([x, y, z]).transpose()
        self._update()

    def set_visible(self, label, visible):
        self.layers[label]['visible'] = visible
        self._update()

    def set_mask(self, label, mask):
        self.layers[label]['mask'] = mask
        self._update()

    def set_errors(self, label, error_lines):
        self.layers[label]['errors'] = error_lines
        self._update()

    def set_vectors(self, label, vectors):
        self.layers[label]['vectors'] = vectors
        self._update()

    def set_draw_arrows(self, label, draw_arrows):
        self.layers[label]['draw_arrows'] = draw_arrows
        self._update()

    def set_size(self, label, size):
        if not np.isscalar(size) and size.ndim > 1:
            raise Exception("size should be a 1-d array")
        self.layers[label]['size'] = size
        self._update()

    def set_color(self, label, rgb):
        if isinstance(rgb, str):
            rgb = ColorConverter().to_rgb(rgb)
        self.layers[label]['color'] = np.asarray(rgb)
        self._update()

    def set_alpha(self, label, alpha):
        self.layers[label]['alpha'] = alpha
        self._update()

    def set_zorder(self, label, zorder):
        self.layers[label]['zorder'] = zorder
        self._update()

    def update_line_width(self, width):
        if self._error_vector_widget:
            self._error_vector_widget.set_data(width=width)

    def _update(self):

        if self._skip_update:
            return

        data = []
        colors = []
        sizes = []
        lines = []
        line_colors = []
        arrows = []
        arrow_colors = []

        for label in sorted(self.layers, key=lambda x: self.layers[x]['zorder']()):

            layer = self.layers[label]

            if not layer['visible'] or layer['data'] is None:
                continue

            input_points = layer['data'].shape[0]
            if layer['mask'] is None:
                n_points = input_points
            else:
                n_points = np.sum(layer['mask'])

            if input_points > 0 and n_points > 0:

                # Data

                if layer['mask'] is None:
                    data.append(layer['data'])
                else:
                    data.append(layer['data'][layer['mask'], :])

                # Colors

                if layer['color'].ndim == 1:
                    rgba = np.hstack([layer['color'], 1])
                    rgba = np.repeat(rgba, n_points).reshape(4, -1).transpose()
                else:
                    if layer['mask'] is None:
                        rgba = layer['color'].copy()
                    else:
                        rgba = layer['color'][layer['mask']]

                rgba[:, 3] *= layer['alpha']

                colors.append(rgba)

                # Sizes

                if np.isscalar(layer['size']):
                    size = np.repeat(layer['size'], n_points)
                else:
                    if layer['mask'] is None:
                        size = layer['size']
                    else:
                        size = layer['size'][layer['mask']]
                sizes.append(size)

                # Error bar and colors
                if layer['errors'] is not None:
                    for error_set in layer['errors']:
                        if layer['mask'] is None:
                            out = error_set
                        else:
                            out = error_set[layer['mask']]
                        out = out.reshape((-1, 3))
                        lines.append(out)
                        line_colors.append(np.repeat(rgba, 2, axis=0))

                if layer['vectors'] is not None:
                    if layer['mask'] is None:
                        out = layer['vectors']
                    else:
                        out = layer['vectors'][layer['mask']]
                    lines.append(out.reshape((-1, 3)))
                    line_colors.append(np.repeat(rgba, 2, axis=0))
                    if layer['draw_arrows']:
                        arrows.append(out)
                        arrow_colors.append(rgba)

        if len(data) == 0:
            self.visible = False
            return
        else:
            self.visible = True

        data = np.vstack(data)
        colors = np.vstack(colors)
        sizes = np.hstack(sizes)

        self.set_data(data, edge_color=colors, face_color=colors, size=sizes)

        if len(lines) == 0:
            if self._error_vector_widget is not None:
                self._error_vector_widget.visible = False
            return
        else:
            if self._error_vector_widget is None:
                widget = Arrow(parent=self, connect="segments")
                widget.set_gl_state(depth_test=False, blend=True,
                                    blend_func=('src_alpha', 'one_minus_src_alpha'))
                self._error_vector_widget = widget
            self._error_vector_widget.visible = True

        lines = np.vstack(lines)
        line_colors = np.vstack(line_colors)
        self._error_vector_widget.set_data(pos=lines, color=line_colors)

        arrows = np.vstack(arrows) if len(arrows) > 0 else np.array([])
        arrow_colors = np.vstack(arrow_colors) if len(arrow_colors) else np.array([])
        self._error_vector_widget.set_data(arrows=arrows)
        self._error_vector_widget.arrow_color = arrow_colors

    def draw(self, *args, **kwargs):
        if len(self.layers) == 0:
            return
        else:
            try:
                super(MultiColorScatter, self).draw(*args, **kwargs)
            except Exception:
                pass


if __name__ == "__main__":  # pragma: nocover

    from vispy import app, scene

    canvas = scene.SceneCanvas(keys='interactive')
    view = canvas.central_widget.add_view()
    view.camera = scene.TurntableCamera(up='z', fov=60)

    x = np.random.random(20)
    y = np.random.random(20)
    z = np.random.random(20)

    multi_scat = MultiColorScatter(parent=view.scene)
    multi_scat.allocate('data')
    multi_scat.set_zorder('data', lambda: 0)
    multi_scat.set_data_values('data', x, y, z)

    multi_scat.allocate('subset1')
    multi_scat.set_mask('subset1', np.random.random(20) > 0.5)
    multi_scat.set_color('subset1', 'red')
    multi_scat.set_zorder('subset1', lambda: 1)

    multi_scat.allocate('subset2')
    multi_scat.set_mask('subset2', np.random.random(20) > 0.5)
    multi_scat.set_color('subset2', 'green')
    multi_scat.set_zorder('subset2', lambda: 2)
    multi_scat.set_alpha('subset2', 0.5)
    multi_scat.set_size('subset2', 20)

    axis = scene.visuals.XYZAxis(parent=view.scene)

    canvas.show()
    app.run()



from ..common.vispy_data_viewer import BaseVispyViewer
from .layer_artist import IsosurfaceLayerArtist
from .layer_style_widget import IsosurfaceLayerStyleWidget
from .viewer_state import Vispy3DIsosurfaceViewerState

from ..common import tools as _tools, selection_tools  # noqa


class VispyIsosurfaceViewer(BaseVispyViewer):

    LABEL = "3D Isosurface Rendering"

    _state_cls = Vispy3DIsosurfaceViewerState
    _layer_style_widget_cls = IsosurfaceLayerStyleWidget

    tools = BaseVispyViewer.tools

    _data_artist_cls = IsosurfaceLayerArtist
    _subset_artist_cls = IsosurfaceLayerArtist

    def add_data(self, data):

        first_layer_artist = len(self._layer_artist_container) == 0

        added = super(VispyIsosurfaceViewer, self).add_data(data)

        if added:
            if first_layer_artist:
                self.state.set_limits(*self._layer_artist_container[0].bbox)
                self._ready_draw = True

        return added



import uuid

import numpy as np

from glue.core.exceptions import IncompatibleAttribute
from glue.utils import categorical_ndarray

from .multi_scatter import MultiColorScatter
from .layer_state import ScatterLayerState
from ..common.layer_artist import VispyLayerArtist

COLOR_PROPERTIES = set(['color_mode', 'cmap_attribute', 'cmap_vmin', 'cmap_vmax', 'cmap', 'color'])
SIZE_PROPERTIES = set(['size_mode', 'size_attribute', 'size_vmin', 'size_vmax',
                       'size_scaling', 'size'])
ERROR_PROPERTIES = set(['xerr_visible', 'yerr_visible', 'zerr_visible',
                        'xerr_attribute', 'yerr_attribute', 'zerr_attribute'])
VECTOR_PROPERTIES = set(['vector_visible', 'vx_attribute', 'vy_attribute', 'vz_attribute',
                         'vector_scaling', 'vector_origin'])
ARROW_PROPERTIES = set(['vector_arrowhead'])
ALPHA_PROPERTIES = set(['alpha'])
DATA_PROPERTIES = set(['layer', 'x_att', 'y_att', 'z_att'])
VISIBLE_PROPERTIES = set(['visible'])


class ScatterLayerArtist(VispyLayerArtist):
    """
    A layer artist to render 3d scatter plots.
    """

    def __init__(self, vispy_viewer, layer=None, layer_state=None):

        super(ScatterLayerArtist, self).__init__(layer)

        self._clip_limits = None

        # Set data caches
        self._marker_data = None
        self._color_data = None
        self._size_data = None

        self.layer = layer or layer_state.layer
        self.vispy_widget = vispy_viewer._vispy_widget

        # TODO: need to remove layers when layer artist is removed
        self._viewer_state = vispy_viewer.state
        self.state = layer_state or ScatterLayerState(layer=self.layer)
        if self.state not in self._viewer_state.layers:
            self._viewer_state.layers.append(self.state)

        # We create a unique ID for this layer artist, that will be used to
        # refer to the layer artist in the MultiColorScatter. We have to do this
        # rather than use self.id because we can't guarantee the latter is
        # unique.
        self.id = str(uuid.uuid4())

        # We need to use MultiColorScatter instance to store scatter plots, but
        # we should only have one per canvas. Therefore, we store the
        # MultiColorScatter instance in the vispy viewer instance.
        if not hasattr(self.vispy_widget, '_multiscat'):
            multiscat = MultiColorScatter()
            multiscat.set_gl_state(depth_test=False,
                                   blend=True,
                                   blend_func=('src_alpha', 'one_minus_src_alpha'))

            self.vispy_widget.add_data_visual(multiscat)
            self.vispy_widget._multiscat = multiscat
            vispy_viewer.options.ui.label_line_width.show()
            vispy_viewer.options.ui.value_line_width.show()

        self._multiscat = self.vispy_widget._multiscat
        self._multiscat.allocate(self.id)
        self._multiscat.set_zorder(self.id, self.get_zorder)

        # Watch for changes in the viewer state which would require the
        # layers to be redrawn
        self._viewer_state.add_global_callback(self._update_scatter)
        self.state.add_global_callback(self._update_scatter)

        self.reset_cache()

    def reset_cache(self):
        self._last_viewer_state = {}
        self._last_layer_state = {}

    @property
    def visual(self):
        return self._multiscat

    def get_zorder(self):
        return self.zorder

    def get_layer_color(self):
        if self.state.color_mode == 'Fixed':
            return self.state.color
        else:
            return self.state.cmap

    def redraw(self):
        """
        Redraw the Vispy canvas
        """
        if self._multiscat is not None:
            self._multiscat._update()
        self.vispy_widget.canvas.update()

    def clear(self):
        """
        Clear the visualization for this layer
        """
        self._multiscat.set_data_values(self.id, [], [], [])
        self._multiscat.set_errors(self.id, [])
        self._multiscat.set_vectors(self.id, None)

    def remove(self):
        """
        Remove the layer artist from the visualization
        """

        if self._multiscat is None:
            return

        self._multiscat.deallocate(self.id)
        self._multiscat = None

        self._viewer_state.remove_global_callback(self._update_scatter)
        self.state.remove_global_callback(self._update_scatter)

    def _update_sizes(self):
        if self.state.size_mode is None:
            pass
        elif self.state.size_mode == 'Fixed':
            self._multiscat.set_size(self.id, self.state.size * self.state.size_scaling)
        else:
            data = self.layer[self.state.size_attribute].ravel()
            if isinstance(data, categorical_ndarray):
                data = data.codes
            if self.state.size_vmax == self.state.size_vmin:
                size = np.ones(data.shape) * 10
            else:
                size = (20 * (data - self.state.size_vmin) /
                        (self.state.size_vmax - self.state.size_vmin))
            size_data = size * self.state.size_scaling
            size_data[np.isnan(data)] = 0.
            self._multiscat.set_size(self.id, size_data)

    def _update_colors(self):
        if self.state.color_mode is None:
            pass
        elif self.state.color_mode == 'Fixed':
            self._multiscat.set_color(self.id, self.state.color)
        else:
            data = self.layer[self.state.cmap_attribute].ravel()
            if isinstance(data, categorical_ndarray):
                data = data.codes
            if self.state.cmap_vmax == self.state.cmap_vmin:
                cmap_data = np.ones(data.shape) * 0.5
            else:
                cmap_data = ((data - self.state.cmap_vmin) /
                             (self.state.cmap_vmax - self.state.cmap_vmin))
            cmap_data = self.state.cmap(cmap_data)
            cmap_data[:, 3][np.isnan(data)] = 0.
            self._multiscat.set_color(self.id, cmap_data)

    def _update_alpha(self):
        self._multiscat.set_alpha(self.id, self.state.alpha)

    def _update_data(self, event=None):

        try:
            x = self.layer[self._viewer_state.x_att].ravel()
            y = self.layer[self._viewer_state.y_att].ravel()
            z = self.layer[self._viewer_state.z_att].ravel()
        except AttributeError:
            return
        except (IncompatibleAttribute, IndexError):
            # The following includes a call to self.clear()
            self.disable_invalid_attributes(self._viewer_state.x_att,
                                            self._viewer_state.y_att,
                                            self._viewer_state.z_att)
            return
        else:
            self._enabled = True

        self._marker_data = np.array([x, y, z]).transpose()

        # We need to make sure we update the sizes and colors in case
        # these were set as arrays, since the size of the data might have
        # changed (in the case of subsets)

        self._multiscat.set_data_values(self.id, x, y, z)

        # Mask points outside the clip limits
        if self._clip_limits is None:
            self._multiscat.set_mask(self.id, None)
        else:

            xmin, xmax, ymin, ymax, zmin, zmax = self._clip_limits

            if xmin <= xmax:
                keep = (x >= xmin) & (x <= xmax)
            else:
                keep = (x <= xmin) & (x >= xmax)

            if ymin <= ymax:
                keep &= (y >= ymin) & (y <= ymax)
            else:
                keep &= (y <= ymin) & (y >= ymax)

            if zmin <= zmax:
                keep &= (z >= zmin) & (z <= zmax)
            else:
                keep &= (z <= zmin) & (z >= zmax)

            self._multiscat.set_mask(self.id, keep)

        self.redraw()

    def _update_errors(self):
        orig_points = self._multiscat.layers[self.id]['data']
        errors = []

        if self.state.xerr_visible:
            line_points = np.tile(orig_points, (1, 2))
            err = self.layer[self.state.xerr_attribute].ravel()
            line_points[:, 0] -= err
            line_points[:, 3] += err
            errors.append(line_points)

        if self.state.yerr_visible:
            line_points = np.tile(orig_points, (1, 2))
            err = self.layer[self.state.yerr_attribute].ravel()
            line_points[:, 1] -= err
            line_points[:, 4] += err
            errors.append(line_points)

        if self.state.zerr_visible:
            line_points = np.tile(orig_points, (1, 2))
            err = self.layer[self.state.zerr_attribute].ravel()
            line_points[:, 2] -= err
            line_points[:, 5] += err
            errors.append(line_points)

        self._multiscat.set_errors(self.id, errors)
        self.redraw()

    def _update_vectors(self):
        if self.state.vector_visible:
            offsets = {'tail': (0, 1), 'middle': (-0.5, 0.5), 'tip': (-1, 0)}
            orig_points = self._multiscat.layers[self.id]['data']
            vector_points = np.zeros((orig_points.shape[0], 6))
            vec_offset = offsets[self.state.vector_origin]
            scale = self.state.vector_scaling
            vx = self.layer[self.state.vx_attribute].ravel()
            vector_points[:, 0] = orig_points[:, 0] + vec_offset[0] * vx * scale
            vector_points[:, 3] = orig_points[:, 0] + vec_offset[1] * vx * scale
            vy = self.layer[self.state.vy_attribute].ravel()
            vector_points[:, 1] = orig_points[:, 1] + vec_offset[0] * vy * scale
            vector_points[:, 4] = orig_points[:, 1] + vec_offset[1] * vy * scale
            vz = self.layer[self.state.vz_attribute].ravel()
            vector_points[:, 2] = orig_points[:, 2] + vec_offset[0] * vz * scale
            vector_points[:, 5] = orig_points[:, 2] + vec_offset[1] * vz * scale
            self._multiscat.set_vectors(self.id, vector_points)
        else:
            self._multiscat.set_vectors(self.id, None)
        self.redraw()

    def _update_visibility(self):
        self._multiscat.set_visible(self.id, self.visible)
        self.redraw()

    def _update_arrow_head(self):
        self._multiscat.set_draw_arrows(self.id, self.state.vector_arrowhead)
        self.redraw()

    @property
    def default_limits(self):
        if self._marker_data is None:
            raise ValueError("Data not yet set")
        dmin = np.nanmin(self._marker_data, axis=0)
        dmax = np.nanmax(self._marker_data, axis=0)
        # TODO: the following can be optimized
        return tuple(np.array([dmin, dmax]).transpose().ravel())

    def set_clip(self, limits):
        self._clip_limits = limits
        self._update_data()

    def _update_scatter(self, force=False, **kwargs):

        if (self._viewer_state.x_att is None or
            self._viewer_state.y_att is None or
            self._viewer_state.z_att is None or
                self.state.layer is None):
            return

        # Figure out which attributes are different from before. Ideally we shouldn't
        # need this but currently this method is called multiple times if an
        # attribute is changed due to x_att changing then hist_x_min, hist_x_max, etc.
        # If we can solve this so that _update_histogram is really only called once
        # then we could consider simplifying this. Until then, we manually keep track
        # of which properties have changed.

        changed = set()

        if not force:

            for key, value in self._viewer_state.as_dict().items():
                if value != self._last_viewer_state.get(key, None):
                    changed.add(key)

            for key, value in self.state.as_dict().items():
                if value != self._last_layer_state.get(key, None):
                    changed.add(key)

        self._last_viewer_state.update(self._viewer_state.as_dict())
        self._last_layer_state.update(self.state.as_dict())

        if force or len(changed & DATA_PROPERTIES) > 0:
            self._update_data()
            force = True

        if force or len(changed & SIZE_PROPERTIES) > 0:
            self._update_sizes()

        if force or len(changed & ERROR_PROPERTIES) > 0:
            self._update_errors()

        if force or len(changed & VECTOR_PROPERTIES) > 0:
            self._update_vectors()

        if force or len(changed & COLOR_PROPERTIES) > 0:
            self._update_colors()

        if force or len(changed & ALPHA_PROPERTIES) > 0:
            self._update_alpha()

        if force or len(changed & VISIBLE_PROPERTIES) > 0:
            self._update_visibility()

        if force or len(changed & ARROW_PROPERTIES) > 0:
            self._update_arrow_head()

    def update(self):
        with self._multiscat.delay_update():
            self._update_scatter(force=True)
        self.redraw()



import numpy as np
from scipy.ndimage import gaussian_filter

from glue.core.data import Subset
from glue.core.exceptions import IncompatibleAttribute

from .layer_state import IsosurfaceLayerState
from ..common.layer_artist import VispyLayerArtist

from vispy.color import BaseColormap

from .multi_iso_visual import MultiIsoVisual

DATA_PROPERTIES = set(['attribute', 'level_low', 'level_high'])
LEVEL_PROPERTIES = set(['step_value'])
COLOR_PROPERTIES = set(['color', 'alpha', 'cmap'])
STEP_PROPERTIES = set(['step'])


# TODO: create colormaps that is prettier
class TransFire(BaseColormap):
    glsl_map = """
    vec4 translucent_grays(int l){
    if (l==1)
        {return $color_0;}
    if (l==2)
        {return $color_1;}
    if (l==3)
        {return $color_2;}
    if (l==4)
        {return $color_3;}
    if (l==5)
        {return $color_4;}
    if (l==6)
        {return $color_5;}
    if (l==7)
        {return $color_6;}
    if (l==8)
        {return $color_7;}
    if (l==9)
        {return $color_8;}
    if (l==10)
        {return $color_9;}

    }
    """
# class AutoCmap(BaseColormap):
#     colors =
#     glsl_map = """
#     vec4 translucent_grays(int l){
#
#     }
#     """
# vec4 translucent_fire(float t) {
#         return vec4(pow(t, 0.5), t, t*t, max(0, t*1.05 - 0.05));
#     }


class IsosurfaceLayerArtist(VispyLayerArtist):
    """
    A layer artist to render isosurfaces.
    """

    def __init__(self, vispy_viewer, layer=None, layer_state=None):

        super(IsosurfaceLayerArtist, self).__init__(layer)

        self._clip_limits = None

        self.layer = layer or layer_state.layer
        self.vispy_widget = vispy_viewer._vispy_widget

        # TODO: need to remove layers when layer artist is removed
        self._viewer_state = vispy_viewer.state
        self.state = layer_state or IsosurfaceLayerState(layer=self.layer)
        if self.state not in self._viewer_state.layers:
            self._viewer_state.layers.append(self.state)

        # self._iso_visual = scene.Isosurface(np.ones((3, 3, 3)), level=0.5, shading='smooth')
        # Create isosurface visual
        self._iso_visual = MultiIsoVisual(np.ones((3, 3, 3)), step=4, relative_step_size=0.5)
        # relative_step_size: ray casting performance, recommond 0.5~1.5)
        self.vispy_widget.add_data_visual(self._iso_visual)

        self._viewer_state.add_global_callback(self._update_volume)
        self.state.add_global_callback(self._update_volume)

        self.reset_cache()

    def reset_cache(self):
        self._last_viewer_state = {}
        self._last_layer_state = {}

    @property
    def bbox(self):
        return (-0.5, self.layer.shape[2] - 0.5,
                -0.5, self.layer.shape[1] - 0.5,
                -0.5, self.layer.shape[0] - 0.5)

    def redraw(self):
        """
        Redraw the Vispy canvas
        """
        self.vispy_widget.canvas.update()

    def clear(self):
        """
        Remove the layer artist from the visualization
        """
        self._iso_visual.parent = None

    def _update_level(self):
        # TODO: set iso clim
        # self._iso_visual.set_data()
        pass

    def _update_step(self):
        # TODO: generate a new color and transparancy scheme based on step num
        self._iso_visual.step = self.state.step
        self.redraw()
        # self._update_color()

    def _update_color(self):
        cmap_data = self.state.cmap(np.linspace(0, 1, 10).tolist())  # self.cmap returns 10 colors
        cmap_data = cmap_data.tolist()
        t = TransFire(colors=cmap_data)
        self._iso_visual.cmap = t
        self.redraw()

    def _update_data(self):

        if isinstance(self.layer, Subset):
            try:
                mask = self.layer.to_mask()
            except IncompatibleAttribute:
                mask = np.zeros(self.layer.data.shape, dtype=bool)
            data = mask.astype(float)
        else:
            data = self.layer[self.state.attribute]

        if self._clip_limits is not None:
            xmin, xmax, ymin, ymax, zmin, zmax = self._clip_limits
            imin, imax = int(np.ceil(xmin)), int(np.ceil(xmax))
            jmin, jmax = int(np.ceil(ymin)), int(np.ceil(ymax))
            kmin, kmax = int(np.ceil(zmin)), int(np.ceil(zmax))
            invalid = -np.inf
            data = data.copy()
            data[:, :, :imin] = invalid
            data[:, :, imax:] = invalid
            data[:, :jmin] = invalid
            data[:, jmax:] = invalid
            data[:kmin] = invalid
            data[kmax:] = invalid

        # self._iso_visual.set_data(np.nan_to_num(data).transpose())
        gaussian_data = gaussian_filter(data/4, 1)

        # TODO: the clim here conflict with set levels
        # self._iso_visual.set_data(
        # np.nan_to_num(gaussian_data),
        # clim=(self.level_low, self.level_high))
        # self._iso_visual.step = self.step

        self._iso_visual.set_data(np.nan_to_num(gaussian_data))
        self.redraw()

    def _update_visibility(self):
        # if self.visible:
        #     self._iso_visual.parent =
        # else:
        #     self._multivol.disable(self.id)
        self.redraw()

    def set_clip(self, limits):
        self._clip_limits = limits
        self._update_data()

    def _update_volume(self, force=False, **kwargs):

        if self.state.attribute is None or self.state.layer is None:
            return

        # Figure out which attributes are different from before. Ideally we shouldn't
        # need this but currently this method is called multiple times if an
        # attribute is changed due to x_att changing then hist_x_min, hist_x_max, etc.
        # If we can solve this so that _update_histogram is really only called once
        # then we could consider simplifying this. Until then, we manually keep track
        # of which properties have changed.

        changed = set()

        if not force:

            for key, value in self._viewer_state.as_dict().items():
                if value != self._last_viewer_state.get(key, None):
                    changed.add(key)

            for key, value in self.state.as_dict().items():
                if value != self._last_layer_state.get(key, None):
                    changed.add(key)

        self._last_viewer_state.update(self._viewer_state.as_dict())
        self._last_layer_state.update(self.state.as_dict())

        if force or len(changed & DATA_PROPERTIES) > 0:
            self._update_data()

        if force or len(changed & LEVEL_PROPERTIES) > 0:
            self._update_level()

        if force or len(changed & COLOR_PROPERTIES) > 0:
            self._update_color()

        if force or len(changed & STEP_PROPERTIES) > 0:
            self._update_step()

    def update(self):
        self._update_volume(force=True)
        self.redraw()



from vispy.color import BaseColormap


def get_translucent_cmap(r, g, b):

    class TranslucentCmap(BaseColormap):
        glsl_map = """
        vec4 translucent_fire(float t) {{
            return vec4({0}, {1}, {2}, t);
        }}
        """.format(r, g, b)

    return TranslucentCmap()



import os

from qtpy import QtWidgets

from glue.utils.qt import load_ui
from echo.qt import autoconnect_callbacks_to_qt

from glue_vispy_viewers.utils import fix_tab_widget_fontsize


class ScatterLayerStyleWidget(QtWidgets.QWidget):

    def __init__(self, layer_artist):

        super(ScatterLayerStyleWidget, self).__init__()

        self.ui = load_ui('layer_style_widget.ui', self,
                          directory=os.path.dirname(__file__))

        fix_tab_widget_fontsize(self.ui.tab_widget)

        self.state = layer_artist.state

        self.layer_artist = layer_artist
        self.layer = layer_artist.layer

        connect_kwargs = {'value_alpha': dict(value_range=(0., 1.)),
                          'value_size_scaling': dict(value_range=(0.1, 10), log=True),
                          'vector_scaling': dict(value_range=(0.1, 10), log=True)}
        self._connections = autoconnect_callbacks_to_qt(self.state, self.ui, connect_kwargs)

        # Set initial values
        self._update_size_mode()
        self._update_color_mode()
        self._update_error_vis()
        self._update_vector_vis()

        self.state.add_callback('color_mode', self._update_color_mode)
        self.state.add_callback('size_mode', self._update_size_mode)
        self.state.add_callback('xerr_visible', self._update_error_vis)
        self.state.add_callback('yerr_visible', self._update_error_vis)
        self.state.add_callback('zerr_visible', self._update_error_vis)
        self.state.add_callback('vector_visible', self._update_vector_vis)

    def _update_size_mode(self, *args):

        if self.state.size_mode == "Fixed":
            self.ui.size_row_2.hide()
            self.ui.combosel_size_attribute.hide()
            self.ui.valuetext_size.show()
        else:
            self.ui.valuetext_size.hide()
            self.ui.combosel_size_attribute.show()
            self.ui.size_row_2.show()

    def _update_color_mode(self, *args):

        if self.state.color_mode == "Fixed":
            self.ui.color_row_2.hide()
            self.ui.color_row_3.hide()
            self.ui.combosel_cmap_attribute.hide()
            self.ui.spacer_color_label.show()
            self.ui.color_color.show()
        else:
            self.ui.color_color.hide()
            self.ui.combosel_cmap_attribute.show()
            self.ui.spacer_color_label.hide()
            self.ui.color_row_2.show()
            self.ui.color_row_3.show()

    def _update_error_vis(self, *args):
        self.ui.combosel_xerr_attribute.setEnabled(self.state.xerr_visible)
        self.ui.combosel_yerr_attribute.setEnabled(self.state.yerr_visible)
        self.ui.combosel_zerr_attribute.setEnabled(self.state.zerr_visible)

    def _update_vector_vis(self, *arg):
        visible = self.state.vector_visible
        self.ui.combosel_vx_attribute.setEnabled(visible)
        self.ui.combosel_vy_attribute.setEnabled(visible)
        self.ui.combosel_vz_attribute.setEnabled(visible)
        self.ui.value_vector_scaling.setEnabled(visible)
        self.ui.combosel_vector_origin.setEnabled(visible)
        self.ui.bool_vector_arrowhead.setEnabled(visible)
